## Filename:      extractor.py
## Description:   extract text from PS and PDF
## Author:        David Dumas <ddumas@math.uic.edu>
## Modified at:   Mon Jan 24 17:18:40 2005
##                
## Copyright (C) 2005  David Dumas
##                
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## See the file COPYING for details.

"""
Extract the human-readable text from a PDF or PS document using the
'pdftotext' and 'pstotext' utilities.

Common ligatures in PDF files are replaced with multi-character
equivalents; PS files are assumed to be from TeX/LaTeX, and ligatures
from the quirky T1 encoding are handled similarly.

The resulting text is return as a unicode string, and optionally
cached in a file encoded via UTF-8.  However, no attempt is made to
find such a cache file before the extraction is performed.
"""

import os,string

class TextExtractor(list):
    def __init__(self,fn,rtype=None,maxlines=2048,text_output=None):
        if rtype:
            self.rtype = rtype
        else:
            self.rtype = fn.split('.')[-1].lower()
        self.text_output = text_output
        self.fn = fn
        self.maxlines = maxlines
        self.extend(self.get_text())
    def joined(self):
        return string.join(self,'')
    def tex_ligature_scrub(self,s):
        s = s.replace('\031','i')
        s = s.replace('\032','j')
        s = s.replace('\033','ff')
        s = s.replace('\034','fi')
        s = s.replace('\035','fl')
        s = s.replace('\036','ffi')
        s = s.replace('\x0c','-')
        return s.decode('LATIN-1','ignore')

    def ligature_scrub(self,s):
        "take a unicode, return a unicode string with some common ligatures replaced by sequences of letters"
        s = s.replace(u'\ufb00','ff')
        s = s.replace(u'\ufb01','fi')
        s = s.replace(u'\ufb02','fl')
        s = s.replace(u'\ufb03','ffi')
        s = s.replace(u'\ufb04','ffl')
        s = s.replace(u'\ufb05','ft')
        s = s.replace(u'\ufb06','st')
        s = s.replace(u'\u00c6','AE')
        s = s.replace(u'\u00e6','ae')
        s = s.replace(u'\u0152','OE')
        s = s.replace(u'\u0153','oe')
        s = s.replace(u'\u0133','ij')
        s = s.replace(u'\u00df','ss')
        return s

    def get_text(self):
        if self.rtype == 'pdf':
            return self.get_text_pdf()
        elif self.rtype == 'ps':
            return self.get_text_ps()
        else:
            return ['']

    def linesfrom(self,p):
         if self.maxlines:
             return p.readlines(self.maxlines)
         else:
             return p.readlines()        

    def output_lines(self,lines):
        if self.text_output:
            f = file(self.text_output,'wb')
            for l in lines:
                f.write(l.encode('UTF-8'))
            f.close()        

    def get_text_pdf(self):
        p = os.popen('pdftotext -nopgbrk -enc UTF-8 "%s" -' % (self.fn))
        lines = self.linesfrom(p)
        slines = [self.ligature_scrub(x.decode('UTF-8','ignore')) for x in lines]
        self.output_lines(slines)
        return slines

    def get_text_ps(self):
        p = os.popen('pstotext -output /dev/stdout "%s"' % (self.fn))
        lines = self.linesfrom(p)
        slines = [self.tex_ligature_scrub(x) for x in lines]
        self.output_lines(slines)
        return slines
